// 
//   Copyright (C) 2005, 2006, 2007 Free Software Foundation, Inc.
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

// 
//

// Stateful live Sprite instance

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "log.h" 
#include "action.h" // for call_method_parsed (call_method_args)
//#include "gnash.h"
#include "render.h"  // for bounds_in_clipping_area()
#include "sprite_instance.h"
#include "movie_definition.h"
#include "MovieClipLoader.h" // @@ temp hack for loading tests
#include "as_value.h"
#include "as_function.h"
#include "text_character_def.h" // @@ temp hack for createTextField exp.
#include "execute_tag.h"
#include "fn_call.h"
#include "Key.h"
#include "movie_root.h"
#include "movie_instance.h"
#include "swf_event.h"
#include "sprite_definition.h"
#include "ActionExec.h"
#include "builtin_function.h"
#include "smart_ptr.h"
#include "VM.h"
#include "Range2d.h" // for getBounds
#include "GnashException.h"
#include "URL.h"
#include "sound_handler.h"

#include <vector>
#include <string>
#include <cmath>

#include <functional> // for mem_fun, bind1st
#include <algorithm> // for for_each
#include <boost/algorithm/string/case_conv.hpp>

// This needs to be included first for NetBSD systems or we get a weird
// problem with pthread_t being defined too many times if we use any
// STL containers.
#ifdef HAVE_PTHREADS
#include <pthread.h>
#endif

namespace gnash {

// Define OLD_GET_MEMBER to thread "default" members within the get_member
// override rather then by using the new 'init_property' call.
// This will likely be removed soon, just here to track possible bugs introduced
// by this change. In particular, the place_and_remove_instane_test is failing 
// for '_x' property, but attachMovieTest is succeeding with the *new* layout
// and failign with the previous.
//
//#define OLD_GET_MEMBER
//

//#define GNASH_DEBUG 1

// Forward declarations
static as_object* getMovieClipInterface();
static void attachMovieClipInterface(as_object& o);
static void attachMovieClipProperties(as_object& o);

// Initialize unnamed instance count
unsigned int sprite_instance::_lastUnnamedInstanceNum=0;

//------------------------------------------------
// Utility funx
//------------------------------------------------

// Execute the actions in the action list, in the given
// environment. The list of action will be consumed
// starting from the first element. When the function returns
// the list should be empty.
void
sprite_instance::execute_actions(sprite_instance::ActionList& action_list)
{
	// action_list may be changed due to actions (appended-to)
	// this loop might be optimized by using an iterator
	// and a final call to .clear() 
	while ( ! action_list.empty() )
	{
		action_buffer* ab = action_list.front();
		action_list.pop_front(); 

		execute_action(*ab);
	}
}

// Wrapper around dynamic_cast to implement user warning.
// To be used by builtin properties and methods.
static sprite_instance*
ensure_sprite(as_object* obj)
{
	sprite_instance* ret = dynamic_cast<sprite_instance*>(obj);
	if ( ! ret )
	{
		throw ActionException("builtin method or gettersetter for sprite objects called against non-sprite instance");
	}
	return ret;
}

static void sprite_play(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	sprite->set_play_state(sprite_instance::PLAY);
}

static void sprite_stop(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	sprite->set_play_state(sprite_instance::STOP);

	// Stop sound stream as well, if such exist
	int stream_id = sprite->get_sound_stream_id();
	if (sprite->get_sound_stream_id() != -1) {
		sound_handler* sh = get_sound_handler();
		if (sh != NULL) sh->stop_sound(stream_id);
		sprite->set_sound_stream_id(-1);
	}

/*	sound_handler* sh = get_sound_handler();
	if (sh != NULL) sh->stop_all_sounds();*/
}

//removeMovieClip() : Void
static void sprite_remove_movieclip(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	sprite_instance* parent = dynamic_cast<sprite_instance*>(sprite->get_parent());
	if (parent)
	{
		parent->remove_display_object(sprite->get_depth(), 0);
	}
}

// attachMovie(idName:String, newName:String,
//             depth:Number [, initObject:Object]) : MovieClip
static void sprite_attach_movie(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	if (fn.nargs < 3 || fn.nargs > 4)
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("attachMovie called with wrong number of arguments"
			" expected 3 to 4, got (%d) - returning undefined",
			fn.nargs);
		);
		return;
	}

	// Get exported resource 
	std::string id_name = fn.arg(0).to_std_string();

	boost::intrusive_ptr<resource> exported = sprite->get_movie_definition()->get_exported_resource(id_name.c_str());
	if ( exported == NULL )
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("attachMovie: '%s': no such exported resource - "
			"returning undefined",
			id_name.c_str());
		);
		return;
	}
	movie_definition* exported_movie = dynamic_cast<movie_definition*>(exported.get());
	if ( ! exported_movie )
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("attachMovie: exported resource '%s' "
			"is not a movie definition (%s) -- "
			"returning undefined",
			id_name.c_str(),
			typeid(*(exported.get())).name());
		);
		return;
	}

	std::string newname = fn.arg(1).to_std_string();

	// should we support negative depths ?
	uint16_t depth_val = uint16_t(fn.arg(2).to_number());

	boost::intrusive_ptr<character> newch = exported_movie->create_character_instance(sprite, depth_val);
	assert( dynamic_cast<sprite_instance*>(newch.get()) );
	assert( newch.get() > (void*)0xFFFF );
	assert(newch->get_ref_count() > 0);

	newch->set_name(newname.c_str());

	// place_character() will set depth on newch
	if ( ! sprite->attachCharacter(*newch, depth_val) )
	{
		log_error("Could not attach character at depth %d", depth_val);
		return;
	}

	newch->setDynamic();

	/// Properties must be copied *after* the call to attachCharacter
	/// because attachCharacter() will reset matrix !!
	if (fn.nargs > 3 ) {
		as_object* initObject = fn.arg(3).to_object();
		if ( initObject ) {
			log_msg("Initializing properties from object");
			newch->copyProperties(*initObject);
		} else {
			// This is actually a valid thing to do,
			// the documented behaviour is to just NOT
			// initializing the properties in this
			// case.
			IF_VERBOSE_MALFORMED_SWF(
			log_aserror("Fourth argument of attachMovie "
				"doesn't cast to an object (%s)",
				fn.arg(3).to_string());
			);
		}
	}

	fn.result->set_as_object(newch.get()); 
}

// attachAudio(id:Object) : Void
static void sprite_attach_audio(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: MovieClip.attachAudio() unimplemented -- "
			"returning undefined");
		warned=true;
	}
}

//createEmptyMovieClip(name:String, depth:Number) : MovieClip
static void sprite_create_empty_movieclip(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	if (fn.nargs != 2)
	{
		if (fn.nargs < 2)
		{
			IF_VERBOSE_ASCODING_ERRORS(
				log_aserror("createEmptyMovieClip needs "
					"2 args, but %d given,"
					" returning undefined.",
					fn.nargs);
			);
			return;
		}
		else
		{
			IF_VERBOSE_ASCODING_ERRORS(
				log_aserror("createEmptyMovieClip takes "
					"2 args, but %d given, discarding"
					" the one in excess.",
					fn.nargs);
			)
		}
	}

	character* ch = sprite->add_empty_movieclip(fn.arg(0).to_string(), int(fn.arg(1).to_number()));
	fn.result->set_as_object(ch);
}

static void sprite_get_depth(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	int n = sprite->get_depth();

	// Macromedia Flash help says: depth starts at -16383 (0x3FFF)
	fn.result->set_int( - (n + 16383 - 1));
}

//swapDepths(target:Object) : Void
static void sprite_swap_depths(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	
	if (fn.nargs < 1)
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("swapDepths needs one arg");
		);
		return;
	}

	sprite_instance* target;
	if (fn.arg(0).is_object() )
	{
		target = (sprite_instance*) fn.arg(0).to_object();
	}
	else
	if (fn.arg(0).is_number() )
	{
		// Macromedia Flash help says: depth starts at -16383 (0x3FFF)
		int target_depth = int(fn.arg(0).to_number()) + 16383 + 1;

		sprite_instance* parent = (sprite_instance*) sprite->get_parent();
		target = (sprite_instance*) parent->get_character_at_depth(target_depth);
	}
	else
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("swapDepths has received invalid arg\n");
    		);
		return;
	}

	if (sprite == NULL || target == NULL)
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("It is impossible to swap NULL character");
		);
		return;
	}

	if (sprite->get_parent() == target->get_parent() && sprite->get_parent() != NULL)
	{
		int target_depth = target->get_depth();
		target->set_depth(sprite->get_depth());
		sprite->set_depth(target_depth);

		sprite_instance* parent = (sprite_instance*) sprite->get_parent();
		parent->swap_characters(sprite, target);
	}
	else
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("Can't swap depth of MovieClips "
			"with different parents");
		);
	}
}

// TODO: wrap the functionality in a sprite_instance method
//       and invoke it from here, this should only be a wrapper
//
//duplicateMovieClip(name:String, depth:Number, [initObject:Object]) : MovieClip
static void sprite_duplicate_movieclip(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	
	if (fn.nargs < 2)
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("duplicateMovieClip needs 2 or 3 args\n");
	    	);
		return;
	}

	// strk question: Would a call to 
	//   sprite->get_movie_defition()->create_instance()
	//   and an add_display_object taking a character_instance
	//   instead of a character *id* be more appropriate ?
	//   (sounds more general)

	// Copy event handlers from sprite
	// We should not copy 'm_action_buffer' since the 'm_method' already contains it
	std::vector<swf_event*>	event_handlers;
	const std::map<event_id, as_value>& sprite_events = sprite->get_event_handlers();
	typedef std::map<event_id, as_value>::const_iterator event_iterator;
	for (event_iterator it = sprite_events.begin(),
		itEnd = sprite_events.end();
		it != itEnd; ++it )
	{
		swf_event* e = new swf_event; // FIXME: who will delete this ?
		e->m_event = it->first;
		e->m_method = it->second;
		event_handlers.push_back(e);
	}

	character* parent = sprite->get_parent();
	sprite_instance* parent_sprite = parent ? parent->to_movie() : NULL;
	character* ch = NULL;
	if (parent_sprite)
	{
		ch = parent_sprite->add_display_object(
			sprite->get_id(),
			fn.arg(0).to_string(),
			event_handlers,
			int(fn.arg(1).to_number()),
			true, // replace if depth is occupied (to drop)
			sprite->get_cxform(),
			sprite->get_matrix(),
			sprite->get_ratio(),
			sprite->get_clip_depth());

		ch->setDynamic();

		// Copy members from initObject
		if (fn.nargs == 3 && ch)
		{
			as_object* initObject = fn.arg(2).to_object();
			if ( initObject ) ch->copyProperties(*initObject);
		}

	}
	fn.result->set_as_object(ch);
}

static void sprite_goto_and_play(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	if (fn.nargs < 1)
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("sprite_goto_and_play needs one arg");
		);
		return;
	}

	// Convert to 0-based
	size_t target_frame = size_t(fn.arg(0).to_number() - 1);
	sprite->goto_frame(target_frame);
	sprite->set_play_state(sprite_instance::PLAY);
}

static void sprite_goto_and_stop(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	if (fn.nargs < 1)
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("sprite_goto_and_stop needs one arg");
		);
		return;
	}

	// Convert to 0-based
	size_t target_frame = size_t(fn.arg(0).to_number() - 1);

	sprite->goto_frame(target_frame);
	sprite->set_play_state(sprite_instance::STOP);
}

static void sprite_next_frame(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	size_t frame_count = sprite->get_frame_count();
	size_t current_frame = sprite->get_current_frame();
	if (current_frame < frame_count)
	{
	    sprite->goto_frame(current_frame + 1);
	}
	sprite->set_play_state(sprite_instance::STOP);
}

static void sprite_prev_frame(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	size_t current_frame = sprite->get_current_frame();
	if (current_frame > 0)
	{
	    sprite->goto_frame(current_frame - 1);
	}
	sprite->set_play_state(sprite_instance::STOP);
}

static void sprite_get_bytes_loaded(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	fn.result->set_int(sprite->get_bytes_loaded());
}

static void sprite_get_bytes_total(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	// @@ horrible uh ?
	fn.result->set_int(sprite->get_bytes_total());
}

// my_mc.loadMovie(url:String [,variables:String]) : Void
static void sprite_load_movie(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	if (fn.nargs < 1) // url
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_msg("Invalid call to MovieClip.loadMove(), "
			"expected 1 or 2 args, got %d - returning undefined",
			fn.nargs);
		);
		return;
	}

	std::string urlstr = fn.arg(0).to_std_string();
	if (urlstr.empty())
	{
		IF_VERBOSE_ASCODING_ERRORS(
		std::stringstream ss; fn.dump_args(ss);
		log_msg("First argument passed to MovieClip.loadMove(%s) "
			"evaluates to an empty string - "
			"returning undefined",
			ss.str().c_str());
		);
		return;
	}
	const URL& baseurl = get_base_url();
	URL url(urlstr, baseurl);

	if (fn.nargs > 1)
	{
		// TODO: implement support for second argument
		log_error("FIXME: second argument of MovieClip.loadMovie(%s, <variables>) "
			"will be discarded (unsupported)", urlstr.c_str());
		//return;
	}

	sprite->loadMovie(url);
	log_warning("MovieClip.loadMovie(%s) - TESTING ", url.str().c_str());


	//log_error("FIXME: %s not implemented yet", __PRETTY_FUNCTION__);
	//moviecliploader_loadclip(fn);
}

// my_mc.unloadMovie() : Void
static void sprite_unload_movie(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	// See http://sephiroth.it/reference.php?id=429

	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: MovieClip.unloadMovie() not implemented yet");
		warned=true;
	}
}

static void sprite_hit_test(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	static bool warned_1_arg = false;
	static bool warned_2_arg = false;
	static bool warned_3_arg = false;

	switch (fn.nargs)
	{
		case 1: // target
		{
			as_value& tgt_val = fn.arg(0);
			character* target = fn.env->find_target(tgt_val);
			if ( ! target )
			{
				IF_VERBOSE_ASCODING_ERRORS(
				log_aserror("Can't find hitTest target %s",
					tgt_val.to_string());
				);
				return;
			}
			if ( ! warned_1_arg ) {
				log_error("FIXME: hitTest(target) unimplemented");
				warned_1_arg=true;
			}
			break;
		}

		case 2: // x, y
		{
			double x = fn.arg(0).to_number();
			double y = fn.arg(1).to_number();
			if ( ! warned_2_arg ) {
				log_error("FIXME: hitTest(%g,%g) unimplemented",
				x,y);
				warned_2_arg=true;
			}
			break;
		}

		case 3: // x, y, shapeFlag
		{
			double x = fn.arg(0).to_number();
			double y = fn.arg(1).to_number();
			bool shapeFlag = fn.arg(2).to_bool();
			if ( ! warned_3_arg ) {
				log_error("FIXME: hitTest(%g,%g,%d) unimplemented",
					x,y,shapeFlag);
				warned_3_arg=true;
			}
			break;
		}

		default:
		{
			IF_VERBOSE_ASCODING_ERRORS(
				log_aserror("hitTest() called with %u args.",
					fn.nargs);
			);
			break;
		}
	}

	return;

}

static void
sprite_create_text_field(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	if (fn.nargs != 6) // name, depth, x, y, width, height
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_msg("createTextField called with %d args, "
			"expected 6 - returning undefined", fn.nargs);
		);
		return;
	}

	if ( ! fn.arg(0).is_string() ) 
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_msg("First argument of createTextField is not a string"
			" - returning undefined");
		);
		return;
	}
	std::string txt_name = fn.arg(0).to_string();

	if ( ! fn.arg(1).is_number() )
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_msg("Second argument of createTextField is not a number"
			" - returning undefined");
		);
		return;
	}
	int txt_depth = int(fn.arg(1).to_number());

	if ( ! fn.arg(2).is_number() ) 
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_msg("Third argument of createTextField is not a number"
			" - returning undefined");
		);
		return;
	}
	float txt_x = fn.arg(2).to_number();

	if ( ! fn.arg(3).is_number() )
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_msg("Fourth argument of createTextField is not a number"
			" - returning undefined");
		);
		return;
	}
	float txt_y = fn.arg(3).to_number();

	if ( ! fn.arg(4).is_number() )
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_msg("Fifth argument of createTextField is not a number"
			" - returning undefined");
		);
		return;
	}
	float txt_width = fn.arg(4).to_number();

	if ( ! fn.arg(5).is_number() ) 
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_msg("Fifth argument of createTextField is not a number"
			" - returning undefined");
		);
		return;
	}
	float txt_height = fn.arg(5).to_number();

	boost::intrusive_ptr<character> txt = sprite->add_textfield(txt_name,
			txt_depth, txt_x, txt_y, txt_width, txt_height);

	// We should return a ref to the newly created
	// TextField here
	fn.result->set_as_object(txt.get());
}

//getNextHighestDepth() : Number
static void
sprite_getNextHighestDepth(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	unsigned int nextdepth = sprite->getNextHighestDepth();
	fn.result->set_double(static_cast<double>(nextdepth));
}

// getURL(url:String, [window:String], [method:String]) : Void
static void
sprite_getURL(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: MovieClip.getURL() not implemented yet");
		warned=true;
	}
}

// getBounds(targetCoordinateSpace:Object) : Object
static void
sprite_getBounds(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	boost::intrusive_ptr<as_object> target;
	if ( fn.nargs > 0 )
	{
		target = fn.arg(0).to_object();
	}

	// TODO: implement 'Range2d<float> character::getBounds(character* ref=NULL)'
	UNUSED(sprite);
	geometry::Range2d<float> bounds(0, 0, 0, 0); //  = sprite->getBounds(target.get());

	double xMin = bounds.getMinX();
	double yMin = bounds.getMinY();
	double xMax = bounds.getMaxX();
	double yMax = bounds.getMaxY();

	boost::intrusive_ptr<as_object> bounds_obj(new as_object());
	bounds_obj->init_member("xMin", as_value(xMin));
	bounds_obj->init_member("yMin", as_value(yMin));
	bounds_obj->init_member("xMax", as_value(xMax));
	bounds_obj->init_member("yMax", as_value(yMax));

	// xMin, xMax, yMin, and yMax
	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: MovieClip.getBounds() not implemented yet (just stubbed)");
		warned=true;
	}

	fn.result->set_as_object(bounds_obj.get());
}

static void
sprite_globalToLocal(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: MovieClip.globalToLocal() not implemented yet");
		warned=true;
	}
}

static void
sprite_endFill(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	sprite->endFill();
}

static void
sprite_lineTo(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	if ( fn.nargs < 2 )
	{
		IF_VERBOSE_ASCODING_ERRORS(
			log_aserror("MovieClip.lineTo() takes two args");
		);
		return;
	}

	float x = PIXELS_TO_TWIPS(fn.arg(0).to_number());
	float y = PIXELS_TO_TWIPS(fn.arg(1).to_number());

	sprite->lineTo(x, y);

}

static void
sprite_moveTo(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	if ( fn.nargs < 2 )
	{
		IF_VERBOSE_ASCODING_ERRORS(
			log_aserror("MovieClip.moveTo() takes two args");
		);
		return;
	}

	float x = PIXELS_TO_TWIPS(fn.arg(0).to_number());
	float y = PIXELS_TO_TWIPS(fn.arg(1).to_number());

	sprite->moveTo(x, y);

}

static void
sprite_lineStyle(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	uint16_t thickness = 0;
	uint8_t r = 0;
	uint8_t g = 0;
	uint8_t b = 0;
	uint8_t a = 255;


	if ( fn.nargs > 0 )
	{
		thickness = uint16_t(PIXELS_TO_TWIPS(uint16_t(fclamp(fn.arg(0).to_number(), 0, 255))));

		if ( fn.nargs > 1 )
		{
			// 2^24 is the max here
			uint32_t rgbval = uint32_t(fclamp(fn.arg(1).to_number(), 0, 16777216));
			r = uint8_t( (rgbval&0xFF0000) >> 16);
			g = uint8_t( (rgbval&0x00FF00) >> 8);
			b = uint8_t( (rgbval&0x0000FF) );

			if ( fn.nargs > 2 )
			{
				float alphaval = fclamp(fn.arg(2).to_number(), 0, 255);
				a = uint8_t(alphaval);
			}
		}

	}


	rgba color(r, g, b, a);
	//log_msg("Color: %s", color.toString().c_str());

	sprite->lineStyle(thickness, color);

}

static void
sprite_curveTo(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	if ( fn.nargs < 4 )
	{
		IF_VERBOSE_ASCODING_ERRORS(
			log_aserror("MovieClip.curveTo() takes four args");
		);
		return;
	}

	float cx = PIXELS_TO_TWIPS(fn.arg(0).to_number());
	float cy = PIXELS_TO_TWIPS(fn.arg(1).to_number());
	float ax = PIXELS_TO_TWIPS(fn.arg(2).to_number());
	float ay = PIXELS_TO_TWIPS(fn.arg(3).to_number());

	sprite->curveTo(cx, cy, ax, ay);

}

static void
sprite_clear(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	sprite->clear();

}

static void
sprite_beginFill(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);

	uint8_t r = 0;
	uint8_t g = 0;
	uint8_t b = 0;
	uint8_t a = 255;

	if ( fn.nargs > 0 )
	{
		// 2^24 is the max here
		uint32_t rgbval = uint32_t(fclamp(fn.arg(0).to_number(), 0, 16777216));
		r = uint8_t( (rgbval&0xFF0000) >> 16);
		g = uint8_t( (rgbval&0x00FF00) >> 8);
		b = uint8_t( (rgbval&0x0000FF) );

	}

	rgba color(r, g, b, a);

	sprite->beginFill(color);

}

static void
sprite_beginGradientFill(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: MovieClip.beginGradientFill() not implemented yet");
		warned=true;
	}
}

// startDrag([lockCenter:Boolean], [left:Number], [top:Number],
// 	[right:Number], [bottom:Number]) : Void`
static void
sprite_startDrag(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: MovieClip.startDrag() not implemented yet");
		warned=true;
	}
}

// stopDrag() : Void
static void
sprite_stopDrag(const fn_call& fn)
{
	sprite_instance* sprite = ensure_sprite(fn.this_ptr);
	UNUSED(sprite);

	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: MovieClip.stopDrag() not implemented yet");
		warned=true;
	}
}

static void
movieclip_ctor(const fn_call& fn)
{
	boost::intrusive_ptr<as_object> clip = new as_object(getMovieClipInterface());
	//attachMovieClipProperties(*clip);
	fn.result->set_as_object(clip.get());
}

#ifndef OLD_GET_MEMBER


static void
sprite_currentframe_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);

	if ( fn.nargs == 0 ) // getter
	{
		fn.result->set_int(ptr->get_current_frame() + 1);
	}
	else // setter
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("Attempt to set read-only property '_currentframe'");
		);
	}

}

static void
sprite_totalframes_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);

	if ( fn.nargs == 0 ) // getter
	{
		fn.result->set_int(ptr->get_frame_count());
	}
	else // setter
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("Attempt to set read-only property '_totalframes'");
		);
	}

}

static void
sprite_framesloaded_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);

	if ( fn.nargs == 0 ) // getter
	{
		fn.result->set_int(ptr->get_loaded_frames());
	}
	else // setter
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("Attempt to set read-only property '_framesloaded'");
		);
	}

}

static void
sprite_target_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);

	if ( fn.nargs == 0 ) // getter
	{
		fn.result->set_string(ptr->getTargetPath().c_str());
	}
	else // setter
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("Attempt to set read-only property '_target'");
		);
	}

}

static void
sprite_name_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);

	if ( fn.nargs == 0 ) // getter
	{
		VM& vm = VM::get();
		const std::string& name = ptr->get_name();
		if ( vm.getSWFVersion() < 6 && name.empty() )
		{
			return;
		} 
		else
		{
			fn.result->set_string(name.c_str());
		}
	}
	else // setter
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("Attempt to set read-only property '_name'");
		);
	}

}

static void
sprite_droptarget_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);
	UNUSED(ptr);

	if ( fn.nargs == 0 ) // getter
	{
		static bool warned = false;
		if ( ! warned )
		{
			log_error("FIXME: MovieClip._droptarget unimplemented");
			warned=true;
		}

		VM& vm = VM::get();
		if ( vm.getSWFVersion() > 5 )
		{
			fn.result->set_string("");
		} 
		else
		{
			return;
		}
	}
	else // setter
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("Attempt to set read-only property '_droptarget'");
		);
	}

}

static void
sprite_url_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);

	if ( fn.nargs == 0 ) // getter
	{
		fn.result->set_string(ptr->get_movie_definition()->get_url().c_str());
	}
	else // setter
	{
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("Attempt to set read-only property '_url'");
		);
	}

}

static void
sprite_highquality_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);
	UNUSED(ptr);

	if ( fn.nargs == 0 ) // getter
	{
		// We don't support quality settings
		fn.result->set_bool(true);
	}
	else // setter
	{
		static bool warned=false;
		if ( ! warned ) {
			log_warning("MovieClip._highquality setting is unsupported");
			warned = true;
		}
	}
}

// TODO: move this to character class, _focusrect seems a generic property
static void
sprite_focusrect_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);
	UNUSED(ptr);

	if ( fn.nargs == 0 ) // getter
	{
		// Is a yellow rectangle visible around a focused movie clip (?)
		// We don't support focuserct settings
		fn.result->set_bool(false);
	}
	else // setter
	{
		static bool warned=false;
		if ( ! warned ) {
			log_warning("MovieClip._focusrect setting is unsupported");
			warned = true;
		}
	}
}

static void
sprite_soundbuftime_getset(const fn_call& fn)
{
	sprite_instance* ptr = ensure_sprite(fn.this_ptr);
	UNUSED(ptr);

	if ( fn.nargs == 0 ) // getter
	{
		// Number of seconds before sound starts to stream.
		fn.result->set_double(0.0);
	}
	else // setter
	{
		static bool warned=false;
		if ( ! warned ) {
			log_warning("MovieClip._soundbuftime setting is unsupported");
			warned = true;
		}
	}
}

#endif // ndef OLD_GET_MEMBER

/// Properties (and/or methods) *inherited* by MovieClip instances
static void
attachMovieClipInterface(as_object& o)
{
	int target_version = o.getVM().getSWFVersion();

	boost::intrusive_ptr<builtin_function> gettersetter;

	// SWF5 or higher
	o.init_member("attachMovie", &sprite_attach_movie);
	o.init_member("play", &sprite_play);
	o.init_member("stop", &sprite_stop);
	o.init_member("gotoAndStop", &sprite_goto_and_stop);
	o.init_member("gotoAndPlay", &sprite_goto_and_play);
	o.init_member("nextFrame", &sprite_next_frame);
	o.init_member("prevFrame", &sprite_prev_frame);
	o.init_member("getBytesLoaded", &sprite_get_bytes_loaded);
	o.init_member("getBytesTotal", &sprite_get_bytes_total);
	o.init_member("loadMovie", &sprite_load_movie);
	o.init_member("unloadMovie", &sprite_unload_movie);
	o.init_member("hitTest", &sprite_hit_test);
	o.init_member("duplicateMovieClip", &sprite_duplicate_movieclip);
	o.init_member("swapDepths", &sprite_swap_depths);
	o.init_member("removeMovieClip", &sprite_remove_movieclip);
	o.init_member("startDrag", &sprite_startDrag);
	o.init_member("stopDrag", &sprite_stopDrag);
	o.init_member("getURL", &sprite_getURL);
	o.init_member("getBounds", &sprite_getBounds);
	o.init_member("globalToLocal", &sprite_globalToLocal);
	if ( target_version  < 6 ) return;

	// SWF6 or higher
	o.init_member("beginFill", &sprite_beginFill);
	o.init_member("beginGradientFill", &sprite_beginGradientFill);
	o.init_member("clear", &sprite_clear);
	o.init_member("curveTo", &sprite_curveTo);
	o.init_member("lineStyle", &sprite_lineStyle);
	o.init_member("lineTo", &sprite_lineTo);
	o.init_member("moveTo", &sprite_moveTo);
	o.init_member("endFill", &sprite_endFill);
	o.init_member("attachAudio", &sprite_attach_audio);
	o.init_member("createTextField", &sprite_create_text_field);
	o.init_member("getDepth", &sprite_get_depth);
	o.init_member("createEmptyMovieClip", &sprite_create_empty_movieclip);
	if ( target_version  < 7 ) return;

	// SWF7 or higher
	o.init_member("getNextHighestDepth", &sprite_getNextHighestDepth);
	if ( target_version  < 8 ) return;

	// TODO: many more methods, see MovieClip class ...

}

/// Properties (and/or methods) attached to every *instance* of a MovieClip 
static void
attachMovieClipProperties(as_object& o)
{
	//int target_version = o.getVM().getSWFVersion();

	boost::intrusive_ptr<builtin_function> gettersetter;

	//
	// Properties (TODO: move to appropriate SWF version section)
	//

	gettersetter = new builtin_function(&character::x_getset, NULL);
	o.init_property("_x", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::y_getset, NULL);
	o.init_property("_y", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::xscale_getset, NULL);
	o.init_property("_xscale", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::yscale_getset, NULL);
	o.init_property("_yscale", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::xmouse_getset, NULL);
	o.init_property("_xmouse", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::ymouse_getset, NULL);
	o.init_property("_ymouse", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::alpha_getset, NULL);
	o.init_property("_alpha", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::visible_getset, NULL);
	o.init_property("_visible", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::width_getset, NULL);
	o.init_property("_width", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::height_getset, NULL);
	o.init_property("_height", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::rotation_getset, NULL);
	o.init_property("_rotation", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::parent_getset, NULL);
	o.init_property("_parent", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_currentframe_getset, NULL);
	o.init_property("_currentframe", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_totalframes_getset, NULL);
	o.init_property("_totalframes", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_framesloaded_getset, NULL);
	o.init_property("_framesloaded", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_target_getset, NULL);
	o.init_property("_target", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_name_getset, NULL);
	o.init_property("_name", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_droptarget_getset, NULL);
	o.init_property("_droptarget", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_url_getset, NULL);
	o.init_property("_url", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_highquality_getset, NULL);
	o.init_property("_highquality", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_focusrect_getset, NULL);
	o.init_property("_focusrect", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&sprite_soundbuftime_getset, NULL);
	o.init_property("_soundbuftime", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::onrollover_getset, NULL);
	o.init_property("onRollOver", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::onrollout_getset, NULL);
	o.init_property("onRollOut", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::onload_getset, NULL);
	o.init_property("onLoad", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::onmouseup_getset, NULL);
	o.init_property("onMouseUp", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::onmousedown_getset, NULL);
	o.init_property("onMouseDown", *gettersetter, *gettersetter);

	gettersetter = new builtin_function(&character::onmousemove_getset, NULL);
	o.init_property("onMouseMove", *gettersetter, *gettersetter);

}

static as_object*
getMovieClipInterface()
{
	static boost::intrusive_ptr<as_object> proto;
	if ( proto == NULL )
	{
		proto = new as_object();
		attachMovieClipInterface(*proto);
		proto->init_member("constructor", &movieclip_ctor); 
	}
	return proto.get();
}

void
movieclip_class_init(as_object& global)
{
	// This is going to be the global MovieClip "class"/"function"
	static boost::intrusive_ptr<builtin_function> cl=NULL;

	if ( cl == NULL )
	{
		cl=new builtin_function(&movieclip_ctor, getMovieClipInterface());
		// replicate all interface to class, to be able to access
		// all methods as static functions
		attachMovieClipInterface(*cl);
		     
	}

	// Register _global.MovieClip
	global.init_member("MovieClip", cl.get());
}


//------------------------------------------------
// sprite_instance helper classes
//------------------------------------------------


/// A DisplayList visitor used to compute its overall height.
//
/// TODO: this seems bogus actually, as it will just return
///       the height of the tallest character in the list !!
///
class HeightFinder {
public:
	float _h;
	HeightFinder(): _h(0) {}
	bool operator() (character* ch)
	{
		float ch_h = ch->get_height();
		if (ch_h > _h) {
			_h = ch_h;
		}
		return true; // keep scanning
	}
	float getHeight() {
		return _h;
	}
};

/// A DisplayList visitor used to compute its overall width.
//
/// TODO: this seems bogus actually, as it will just return
///       the width of the widest character in the list !!
///
class WidthFinder {
public:
	float _w;
	WidthFinder(): _w(0) {}
	bool operator() (character* ch) 
	{
		float ch_w = ch->get_width();
		if (ch_w > _w) {
			_w = ch_w;
		}
		return true; // keep scanning
	}
	float getWidth() {
		return _w;
	}
};

/// A DisplayList visitor used to extract script characters
//
/// Script characters are characters created or transformed
/// by ActionScript
///
class ScriptObjectsFinder {
	std::vector<character*>& _dynamicChars;
	std::vector<character*>& _staticChars;
public:
	ScriptObjectsFinder(std::vector<character*>& dynamicChars,
			std::vector<character*>& staticChars)
		:
		_dynamicChars(dynamicChars),
		_staticChars(staticChars)
	{}

	bool operator() (character* ch) 
	{
		// TODO: Are script-transformed object to be kept ?
		//       Need a testcase for this
		//if ( ! ch->get_accept_anim_moves() )
		if ( ch->isDynamic() )
		{
			_dynamicChars.push_back(ch);
		}
		else
		{
			_staticChars.push_back(ch);
		}
		return true; // keep scanning
	}
};

/// A DisplayList visitor used to extract static characters
//
/// Static characters are characters instantiaced trough display-list SWF tags
///
class StaticObjectsFinder {
	std::vector<character*>& _chars;
public:
	StaticObjectsFinder(std::vector<character*>& chars)
		:
		_chars(chars)
	{}

	bool operator() (character* ch) 
	{
		//if ( ch->get_accept_anim_moves() )
		if ( ! ch->isDynamic() )
		{
			_chars.push_back(ch);
		}
		return true; // keep scanning
	}
};

/// A DisplayList visitor used to extract all characters
//
/// Script characters are characters created or transformed
/// by ActionScript
///
class CharactersExtractor {
	std::vector<character*>& _chars;
public:
	CharactersExtractor(std::vector<character*>& chars)
		:
		_chars(chars)
	{}

	bool operator() (character* ch) 
	{
		_chars.push_back(ch);
		return true; // keep scanning
	}
};


//------------------------------------------------
// sprite_instance
//------------------------------------------------

sprite_instance::sprite_instance(
		movie_definition* def, movie_instance* r,
		character* parent, int id)
	:
	character(parent, id),
	m_mouse_state(UP),
	m_root(r),
	m_display_list(),
	oldDisplayList(),
	_drawable(new DynamicShape),
	_drawable_inst(_drawable->create_character_instance(this, 0)),
	m_action_list(),
	m_goto_frame_action_list(),
	m_play_state(PLAY),
	m_current_frame(0),
	m_update_frame(true),
	m_has_looped(false),
	m_init_actions_executed(),
	m_as_environment(),
	m_has_keypress_event(false),
	m_has_mouse_event(false),
	_text_variables(),
	m_sound_stream_id(-1),
	_target(), // don't initialize now, as parent can be assigned later
	m_def(def),
	m_on_event_load_called(false)
{
	assert(m_def != NULL);
	assert(m_root != NULL);

	set_prototype(getMovieClipInterface());
			
	//m_root->add_ref();	// @@ circular!
	m_as_environment.set_target(this);

	// Initialize the flags for init action executed.
	m_init_actions_executed.assign(m_def->get_frame_count(), false);

	// TODO: have the 'MovieClip' constructor take care of this !
	attachMovieClipProperties(*this);

}

sprite_instance::~sprite_instance()
{

	if (m_has_keypress_event)
	{
		_vm.getRoot().remove_keypress_listener(this);
	}

	if (m_has_mouse_event)
	{
		_vm.getRoot().remove_mouse_listener(this);
	}

	m_display_list.clear();
}

character* sprite_instance::get_character_at_depth(int depth)
{
	return m_display_list.get_character_at_depth(depth);
}

// Set *val to the value of the named member and
// return true, if we have the named member.
// Otherwise leave *val alone and return false.
bool sprite_instance::get_member(const std::string& name, as_value* val)
{
	// FIXME: use addProperty interface for these !!
	if ( name == "_root" )
	{
		// TODO: handle lockroot
		val->set_as_object( VM::get().getRoot().get_root_movie() );
		return true;
	}
	if ( name == "_level0" )
	{
		// TODO: handle _level# (any level)
		val->set_as_object( VM::get().getRoot().get_root_movie() );
		return true;
	}
	if ( name == "this" )
	{
		val->set_as_object( this );
		return true;
	}

#ifdef OLD_GET_MEMBER
	// FIXME: use addProperty interface for these !!
	as_standard_member std_member = get_standard_member(name);
	switch (std_member)
	{
	default:
	case M_INVALID_MEMBER:
	    break;
	case M_X:
	    //if (name == "_x")
	{
	    matrix	m = get_matrix();
	    val->set_double(TWIPS_TO_PIXELS(m.m_[0][2]));
	    return true;
	}
	case M_Y:
	    //else if (name == "_y")
	{
	    matrix	m = get_matrix();
	    val->set_double(TWIPS_TO_PIXELS(m.m_[1][2]));
	    return true;
	}
	case M_XSCALE:
	    //else if (name == "_xscale")
	{
	    matrix m = get_matrix();	// @@ or get_world_matrix()?  Test this.
	    float xscale = m.get_x_scale();
	    val->set_double(xscale * 100);		// result in percent
	    return true;
	}
	case M_YSCALE:
	    //else if (name == "_yscale")
	{
	    matrix m = get_matrix();	// @@ or get_world_matrix()?  Test this.
	    float yscale = m.get_y_scale();
	    val->set_double(yscale * 100);		// result in percent
	    return true;
	}
	case M_CURRENTFRAME:
	    //else if (name == "_currentframe")
	{
	    val->set_int(m_current_frame + 1);
	    return true;
	}
	case M_TOTALFRAMES:
	    //else if (name == "_totalframes")
	{
	    // number of frames.  Read only.
	    val->set_int(m_def->get_frame_count());
	    return true;
	}
	case M_ALPHA:
	    //else if (name == "_alpha")
	{
	    // Alpha units are in percent.
	    val->set_double(get_cxform().m_[3][0] * 100.f);
	    return true;
	}
	case M_VISIBLE:
	    //else if (name == "_visible")
	{
	    val->set_bool(get_visible());
	    return true;
	}
	case M_WIDTH: // _width
	{
	    // Verified using samples/test_rotation.swf
	    val->set_double(TWIPS_TO_PIXELS(get_width()));
	    return true;
	}
	case M_HEIGHT: // _height
	{
	    // Verified using samples/test_rotation.swf
	    val->set_double(TWIPS_TO_PIXELS(get_height()));
	    return true;
	}
	case M_ROTATION:
	    //else if (name == "_rotation")
	{
	    // Verified against Macromedia player using samples/test_rotation.swf
	    float	angle = get_matrix().get_rotation();

	    // Result is CLOCKWISE DEGREES, [-180,180]
	    angle *= 180.0f / float(M_PI);

	    val->set_double(angle);
	    return true;
	}
	case M_PARENT:
	{
		if (m_parent==NULL)
		// _parent is undefined for root movies
		{
			return false;
		}
		else
		{
			assert(dynamic_cast<as_object*>(get_parent()));
			val->set_as_object(static_cast<as_object*>(get_parent()));
			return true;
		}
	}
	case M_FRAMESLOADED:
	    //else if (name == "_framesloaded")
	{
	    val->set_int(m_def->get_loading_frame());
	    return true;
	}

	case M_XMOUSE:
	    //else if (name == "_xmouse")
	{
	    // Local coord of mouse IN PIXELS.
	    int	x, y, buttons;
	    _vm.getRoot().get_mouse_state(x, y, buttons);

	    matrix	m = get_world_matrix();

	    point	a(PIXELS_TO_TWIPS(x), PIXELS_TO_TWIPS(y));
	    point	b;
			
	    m.transform_by_inverse(&b, a);

	    val->set_double(TWIPS_TO_PIXELS(b.m_x));
	    return true;
	}

	case M_YMOUSE:
	    //else if (name == "_ymouse")
	{
	    // Local coord of mouse IN PIXELS.
	    int	x, y, buttons;
	    _vm.getRoot().get_mouse_state(x, y, buttons);

	    matrix	m = get_world_matrix();

	    point	a(PIXELS_TO_TWIPS(x), PIXELS_TO_TWIPS(y));
	    point	b;
			
	    m.transform_by_inverse(&b, a);

	    val->set_double(TWIPS_TO_PIXELS(b.m_y));
	    return true;
	}
	/// FIXME: use a contextual 'target' member
	case M_TARGET:
	    //else if (name == "_target")
	{
	    // Full path to this object; e.g. "/_level0/sprite1/sprite2/ourSprite"
	    val->set_string(getTargetPath().c_str());
	    return true;
	}
	case M_NAME:
	    //else if (name == "_name")
	{
	    if ( _vm.getSWFVersion() < 6 && get_name().empty() )
		val->set_undefined();
	    else
	    	val->set_string(get_name().c_str());

	    return true;
	}

	case M_DROPTARGET:
	    //else if (name == "_droptarget")
	{
	    // Absolute path in slash syntax where we were last dropped (?)
	    // @@ TODO
		static bool warned = false;
		if ( ! warned ) {
			log_error("FIXME: MovieClip._droptarget unimplemented");
			warned=true;
		}

	    if ( _vm.getSWFVersion() > 5 )
	    	val->set_string("");
	    else
		val->set_undefined();
	    return true;
	}
	case M_URL:
	    //else if (name == "_url")
	{
		// A sprite's url is the url this
		// sprite has been "downloaded" from.
		// If this is an ActionScript-created sprite
		// we might use the actions creating
		// it as defining where was it "downloaded" from.
		//
	    val->set_string(m_def->get_url().c_str()); 
	    return true;
	}
	case M_ONROLLOVER:
	{
		return get_event_handler(event_id::ROLL_OVER, val);
	}
	case M_ONROLLOUT:
	{
		return get_event_handler(event_id::ROLL_OUT, val);
	}
	case M_HIGHQUALITY:
	    //else if (name == "_highquality")
	{
	    // Whether we're in high quality mode or not.
	    val->set_bool(true);
	    return true;
	}
	case M_FOCUSRECT:
	    //else if (name == "_focusrect")
	{
	    // Is a yellow rectangle visible around a focused movie clip (?)
	    val->set_bool(false);
	    return true;
	}
	case M_SOUNDBUFTIME:
	    //else if (name == "_soundbuftime")
	{
	    // Number of seconds before sound starts to stream.
	    val->set_double(0.0);
	    return true;
	}
	case M_ONLOAD:
	{
	    if (m_as_environment.get_member(std::string(name.c_str()), val))
		{
		    return true;
		}
	    // Optimization: if no hit, don't bother looking in the display list, etc.
	    return false;
	}

	}	// end switch
#endif // def OLD_GET_MEMBER

	// Try variables.
	if ( m_as_environment.get_member(name, val) )
	{
	    return true;
	}

	// Try object members, BEFORE display list items!
	// (see testcase VarAndCharClash.swf in testsuite/misc-ming.all)
	//
	// TODO: simplify the next line when get_member_default takes
	//       a std::string
	if ( get_member_default(name.c_str(), val) )
	{

// ... trying to be useful to Flash coders ...
// The check should actually be performed before any return
// prior to the one due to a match in the DisplayList.
// It's off by default anyway, so not a big deal.
// See bug #18457
//#define CHECK_FOR_NAME_CLASHES 1
#ifdef CHECK_FOR_NAME_CLASHES
		IF_VERBOSE_ASCODING_ERRORS(
		if (  m_display_list.get_character_by_name_i(name) )
		{
			log_aserror("A sprite member (%s) clashes with "
					"the name of an existing character "
					"in its display list! "
					"The member will hide the "
					"character.", name.c_str());
		}
		);
#endif

		return true;
	}


	// Try items on our display list.
	character* ch = m_display_list.get_character_by_name_i(name);
	if (ch)
	{
	    // Found object.
	    val->set_as_object(static_cast<as_object*>(ch));
	    return true;
	}

	// Try textfield variables
	edit_text_character* etc = get_textfield_variable(name);
	if ( etc )
	{
	    	val->set_string(etc->get_text_value());
		return true;
	}

	return false;

}

// Take care of this frame's actions.
void sprite_instance::do_actions()
{
	testInvariant();

	IF_VERBOSE_ACTION(
		log_action("Executing %u actions in frame %u/%u"
			" of sprite %s",
			m_action_list.size(),
			m_current_frame+1,
			m_def->get_frame_count(), getTargetPath().c_str());
	);

	execute_actions(m_action_list);
	assert(m_action_list.empty());

	testInvariant();
}

size_t
sprite_instance::get_frame_number(const as_value& frame_spec) const
{
	size_t frame_number;

	// Figure out what frame to call.
	if (frame_spec.is_string())
	{
		if (m_def->get_labeled_frame(frame_spec.to_string(), &frame_number) == false)
		{
			// Try converting to integer.
			frame_number = (size_t)frame_spec.to_number();
		}
	}
	else
	{
		// convert from 1-based to 0-based
		frame_number = (size_t) frame_spec.to_number() - 1;
	}

	return frame_number;
}

/// Execute the actions for the specified frame. 
//
/// The frame_spec could be an integer or a string.
///
void sprite_instance::call_frame_actions(const as_value& frame_spec)
{
	size_t	frame_number = get_frame_number(frame_spec);

	if (frame_number >= m_def->get_frame_count())
	{
		// No dice.
		IF_VERBOSE_ASCODING_ERRORS(
		log_aserror("call_frame('%s') -- unknown frame", frame_spec.to_string());
		);
		return;
	}

	// Take not of iterator to last element
	ActionList::iterator top_iterator = m_action_list.end();
	--top_iterator; // now points to last element in *current* list

#ifndef NDEBUG
	size_t original_size = m_action_list.size();
#endif

	// Set the current sound_stream_id to -1, meaning that no stream are
	// active. If there are an active stream it will be updated while
	// executing the execute_tags.
	set_sound_stream_id(-1);

	// Execute the execute_tag actions

	const PlayList& playlist = m_def->get_playlist(frame_number);
	for (size_t i=0, n=playlist.size(); i<n; ++i)
	{
		execute_tag*	e = playlist[i];
		if (e->is_action_tag())
		{
			e->execute(this);
		}
	}

	// Execute any new actions triggered by the tag,
	// leaving existing actions to be executed.

	++top_iterator; // now points to one past last of *previous* list
	ActionList::const_iterator it = top_iterator;
	while (it != m_action_list.end())
	{
		execute_action(*(*it));
		++it;
	}
	m_action_list.erase(top_iterator, m_action_list.end());

	assert(m_action_list.size() == original_size);
}

character* sprite_instance::add_empty_movieclip(const char* name, int depth)
{
	cxform color_transform;
	matrix matrix;

	// empty_sprite_def will be deleted during deliting sprite
	sprite_definition* empty_sprite_def = new sprite_definition(get_movie_definition(), NULL);

	sprite_instance* sprite = new sprite_instance(empty_sprite_def, m_root, this, 0);
	sprite->set_name(name);
	sprite->setDynamic();

	m_display_list.place_character(
		sprite,
		depth,
		color_transform,
		matrix,
		0.0f,
		0); 

	return sprite;
}

boost::intrusive_ptr<character>
sprite_instance::add_textfield(const std::string& name, int depth, float x, float y, float width, float height)
{
	matrix txt_matrix;

	// Do I need the smart_ptr.here ?
	boost::intrusive_ptr<text_character_def> txt = new text_character_def(get_movie_definition());
	boost::intrusive_ptr<character> txt_char = txt->create_character_instance(this, 0);

	txt_char->set_name(name.c_str());
	txt_char->setDynamic();

	// Here we should set the proper matrix using x,y width and height

	// set _x
	txt_matrix.m_[0][2] = infinite_to_fzero(PIXELS_TO_TWIPS(x));
	// set _y
	txt_matrix.m_[1][2] = infinite_to_fzero(PIXELS_TO_TWIPS(y));
	// set _width
	txt_matrix.m_[0][0] = infinite_to_fzero(PIXELS_TO_TWIPS(width));
	// set _height
	txt_matrix.m_[1][1] = infinite_to_fzero(PIXELS_TO_TWIPS(height));

	// Here we add the character to the displayList.

	m_display_list.place_character(
		txt_char.get(),
		depth,
		cxform(),
		txt_matrix,
		0.0f,
		0);

	static bool warned = false;
	if ( ! warned )
	{
		log_error("FIXME: %s unfinished", __PRETTY_FUNCTION__);
		warned = true;
	}

	return txt_char;
}

void sprite_instance::clone_display_object(const std::string& name,
	const std::string& newname, uint16_t depth)
{
//            GNASH_REPORT_FUNCTION;

    character* ch = m_display_list.get_character_by_name(name);
    if (ch)
	{
	    std::vector<swf_event*>	dummy_event_handlers;

	    add_display_object(
		ch->get_id(),
		newname.c_str(),
		dummy_event_handlers,
		depth,
		true, // replace if depth is occupied (to drop)
		ch->get_cxform(),
		ch->get_matrix(),
		ch->get_ratio(),
		ch->get_clip_depth());
	    // @@ TODO need to duplicate ch's event handlers, and presumably other members?
	    // Probably should make a character::clone() function to handle this.
	}
}

void sprite_instance::remove_display_object(const tu_string& name_tu)
{
//	    GNASH_REPORT_FUNCTION;

	std::string name(name_tu.c_str());

	character* ch = m_display_list.get_character_by_name(name);
	if (ch)
	{
	    // @@ TODO: should only remove movies that were created via clone_display_object --
	    // apparently original movies, placed by anim events, are immune to this.
	    remove_display_object(ch->get_depth(), ch->get_id());
	}
}

bool
sprite_instance::on_event(const event_id& id)
{
	testInvariant();

	bool called = false;
			
	// First, check for built-in event handler.
	{
		as_value	method;
		if (get_event_handler(id, &method))
		{
			// Dispatch.
			call_method0(method, &m_as_environment, this);

			called = true;
			// Fall through and call the function also, if it's defined!
			// (@@ Seems to be the behavior for mouse events; not tested & verified for
			// every event type.)
		}
	}

	// If we called a built-in event handler
	// we stop here. If we don't do this,
	// MOUSE_UP, MOUSE_DOWN and MOUSE_MOVE will
	// be called twice !
	// Don't know if this is generally valid for every
	// kind of event, but we do have testcases for this one...
	if ( called ) return true;

	// Check for member function.
	{
		// In ActionScript 2.0, event method names are CASE SENSITIVE.
		// In ActionScript 1.0, event method names are CASE INSENSITIVE.
		// TODO: move to get_function_name directly ?
		std::string method_name = id.get_function_name();
		if ( _vm.getSWFVersion() < 7 )
		{
			boost::to_lower(method_name, _vm.getLocale());
		}

		if (method_name.length() > 0)
		{
			as_value	method;
			if (get_member(method_name, &method) && ! method.is_undefined() )
			{
				call_method0(method, &m_as_environment, this);
				called = true;
			}
		}
	}

	testInvariant();

	return called;
}

character*
sprite_instance::get_relative_target(const std::string& name)
{
	character* ch = get_relative_target_common(name);

	if ( ! ch )
	{
		// See if we have a match on the display list.
		return m_display_list.get_character_by_name(name);
	}

	return ch; // possibly NULL
}

void sprite_instance::set_member(const std::string& name,
		const as_value& val)
{
#ifdef DEBUG_DYNTEXT_VARIABLES
log_msg("sprite[%p]::set_member(%s, %s)", (void*)this, name.c_str(), val.to_string());
#endif

#ifdef OLD_GET_MEMBER
	as_standard_member	std_member = get_standard_member(name);
	switch (std_member)
	{
		default:
		case M_INVALID_MEMBER:
		    break;
		case M_X:
		    //if (name == "_x")
		{
		    matrix	m = get_matrix();
		    m.m_[0][2] = infinite_to_fzero(PIXELS_TO_TWIPS(val.to_number()));
		    set_matrix(m);

                    transformedByScript(); // m_accept_anim_moves = false; 

		    return;
		}
		case M_Y:
		    //else if (name == "_y")
		{
		    matrix	m = get_matrix();
		    m.m_[1][2] = infinite_to_fzero(PIXELS_TO_TWIPS(val.to_number()));
		    set_matrix(m);

                    transformedByScript(); // m_accept_anim_moves = false; 

		    return;
		}
		case M_XSCALE:
		    //else if (name == "_xscale")
		{
		    matrix	m = get_matrix();

                    double scale_percent = val.to_number();

                    // Handle bogus values
                    if (isnan(scale_percent))
                    {
			IF_VERBOSE_ASCODING_ERRORS(
			log_aserror("Attempt to set _xscale to %g, refused",
                            scale_percent);
			);
                        return;
                    }
                    else if (scale_percent < 0 )
                    {
			IF_VERBOSE_ASCODING_ERRORS(
			log_aserror("Attempt to set _xscale to %g, use 0",
                            scale_percent);
			);
                        scale_percent = 0;
                    }

                    // input is in percent
		    float scale = (float)scale_percent/100.f;

		    // Decompose matrix and insert the desired value.
		    float	x_scale = scale;
		    float	y_scale = m.get_y_scale();
		    float	rotation = m.get_rotation();
		    m.set_scale_rotation(x_scale, y_scale, rotation);

		    set_matrix(m);
		    transformedByScript(); // m_accept_anim_moves = false; 
		    return;
		}
		case M_YSCALE:
		    //else if (name == "_yscale")
		{
		    matrix	m = get_matrix();

                    double scale_percent = val.to_number();

                    // Handle bogus values
                    if (isnan(scale_percent))
                    {
			IF_VERBOSE_ASCODING_ERRORS(
			log_aserror("Attempt to set _yscale to %g, refused",
                            scale_percent);
			);
                        return;
                    }
                    else if (scale_percent < 0 )
                    {
			IF_VERBOSE_ASCODING_ERRORS(
			log_aserror("Attempt to set _yscale to %g, use 0",
                            scale_percent);
			);
                        scale_percent = 0;
                    }

                    // input is in percent
		    float scale = (float)scale_percent/100.f;

		    // Decompose matrix and insert the desired value.
		    float	x_scale = m.get_x_scale();
		    float	y_scale = scale;
		    float	rotation = m.get_rotation();
		    m.set_scale_rotation(x_scale, y_scale, rotation);

		    set_matrix(m);
                    transformedByScript(); // m_accept_anim_moves = false; 
		    return;
		}
		case M_ALPHA:
		    //else if (name == "_alpha")
		{
		    // Set alpha modulate, in percent.
		    cxform	cx = get_cxform();
		    cx.m_[3][0] = infinite_to_fzero(val.to_number()) / 100.f;
		    set_cxform(cx);
                    transformedByScript(); // m_accept_anim_moves = false; 
		    return;
		}
		case M_VISIBLE:
		    //else if (name == "_visible")
		{
		    set_visible(val.to_bool());
                    transformedByScript(); // m_accept_anim_moves = false; 
		    return;
		}
		case M_WIDTH:
		    //else if (name == "_width")
		{
		    // @@ tulrich: is parameter in world-coords or local-coords?
		    matrix	m = get_matrix();
		    m.m_[0][0] = infinite_to_fzero(PIXELS_TO_TWIPS(val.to_number()));
		    float w = get_width();
		    if (fabsf(w) > 1e-6f)
			{
			    m.m_[0][0] /= w;
			}
		    set_matrix(m);
                    transformedByScript(); // m_accept_anim_moves = false; 
		    return;
		}
		case M_HEIGHT:
		    //else if (name == "_height")
		{
		    // @@ tulrich: is parameter in world-coords or local-coords?
		    matrix	m = get_matrix();
		    m.m_[1][1] = infinite_to_fzero(PIXELS_TO_TWIPS(val.to_number()));
		    float h = get_width(); // FIXME: shoudln't this be get_height ??
		    if (fabsf(h) > 1e-6f)
			{
			    m.m_[1][1] /= h;
			}
		    set_matrix(m);
                    transformedByScript(); // m_accept_anim_moves = false; 
		    return;
		}
		case M_ROTATION:
		    //else if (name == "_rotation")
		{
		    matrix	m = get_matrix();

		    // Decompose matrix and insert the desired value.
		    float	x_scale = m.get_x_scale();
		    float	y_scale = m.get_y_scale();
		    float	rotation = (float) val.to_number() * float(M_PI) / 180.f;	// input is in degrees
		    m.set_scale_rotation(x_scale, y_scale, rotation);

		    set_matrix(m);
                    transformedByScript(); // m_accept_anim_moves = false; 
		    return;
		}
		case M_ONROLLOVER:
		{
			set_event_handler(event_id::ROLL_OVER, val);
			return;
		}
		case M_ONROLLOUT:
		{
			set_event_handler(event_id::ROLL_OUT, val);
			return;
		}
		case M_HIGHQUALITY:
		    //else if (name == "_highquality")
		{
		    // @@ global { 0, 1, 2 }
//				// Whether we're in high quality mode or not.
//				val->set(true);
		    return;
		}
		case M_FOCUSRECT:
		    //else if (name == "_focusrect")
		{
//				// Is a yellow rectangle visible around a focused movie clip (?)
//				val->set(false);
		    return;
		}
		case M_SOUNDBUFTIME:
		    //else if (name == "_soundbuftime")
		{
		    // @@ global
//				// Number of seconds before sound starts to stream.
//				val->set(0.0);
		    return;
		}
	}	// end switch
#endif // def OLD_GET_MEMBER

#ifdef DEBUG_DYNTEXT_VARIABLES
log_msg(" not a standard member");
#endif


#if 0 // we'd need a testcase for this, my "short" tests show
      // that a textfield variable cannot be assigned to like this

	// Not a built-in property.  See if we have a
	// matching edit_text character in our display
	// list.
	bool text_val = val.is_string() || val.is_number();
	if (text_val)
	{
			// CASE INSENSITIVE compare. 
			// In ActionScript 2.0, this must change
			// to CASE SENSITIVE!!!
			character* ch = m_display_list.get_character_by_name_i( name.c_str());
			if ( ch ) // item found
			{
				const char* text = val.to_string();
				ch->set_text_value(text);
				return;
			}
	}

#ifdef DEBUG_DYNTEXT_VARIABLES
	log_msg(" not a character");
#endif
#endif


	// Try textfield variables
	//
	// FIXME: Turn textfield variables into Getter/Setters (Properties)
	//        so that set_member_default will do this automatically.
	//        The problem is that setting a TextVariable named after
	//        a builtin property will prevent *any* setting for the
	//        property (ie: have a textfield use _x as variable name and
	//        be scared)
	//
	edit_text_character* etc = get_textfield_variable(name.c_str());
	if ( etc )
	{
#ifdef DEBUG_DYNTEXT_VARIABLES
log_msg(" it's a Text Variable!");
#endif
		etc->set_text_value(val.to_string());
	}
#ifdef DEBUG_DYNTEXT_VARIABLES
	else
	{
log_msg(" it's NOT a Text Variable!");
	}
#endif

	// If that didn't work call the default set_member
	set_member_default(name, val);


	// If that didn't work, set a variable within this environment.
	// TODO: check if we broke anything with this!
	//if ( ! t ) m_as_environment.set_member(name.c_str(), val);
}

const char* sprite_instance::get_variable(const char* path_to_var) const
{
    assert(m_parent == NULL);	// should only be called on the root movie.

    std::string path(path_to_var);

    // NOTE: this is static so that the string
    // value won't go away after we return!!!
    // It'll go away during the next call to this
    // function though!!!  NOT THREAD SAFE!
    static as_value	val;

    val = m_as_environment.get_variable(path);

    return val.to_string();	// ack!
}

void sprite_instance::set_variable(const char* path_to_var,
		const wchar_t* new_value)
{
	if (path_to_var == NULL)
	{
		log_error("NULL path_to_var passed to set_variable()\n");
		return;
	}
	if (new_value == NULL)
	{
		log_error("NULL passed to set_variable('%s',"
			" NULL)\n", path_to_var);
		return;
	}

	// should only be called on the root movie.
	assert(m_parent == NULL);

	std::string path(path_to_var);
	as_value val(new_value);

	m_as_environment.set_variable(path, val);
}

void sprite_instance::set_variable(const char* path_to_var,
		const char* new_value)
{
	    assert(m_parent == NULL);	// should only be called on the root movie.

	    if (path_to_var == NULL)
		{
		    log_error("NULL path_to_var passed to set_variable()\n");
		    return;
		}
	    if (new_value == NULL)
		{
		    log_error("NULL passed to set_variable('%s', NULL)\n", path_to_var);
		    return;
		}

	    std::string path(path_to_var);
	    as_value val(new_value);

	    m_as_environment.set_variable(path, val);
}

void sprite_instance::advance_sprite(float delta_time)
{
	//GNASH_REPORT_FUNCTION;

	// mouse drag.
	character::do_mouse_drag();

	if (m_on_event_load_called)
	{
		on_event(event_id::ENTER_FRAME);
	}

#ifdef GNASH_DEBUG
	size_t frame_count = m_def->get_frame_count();

	log_msg("sprite '%s' ::advance_sprite is at frame %u/%u "
		"- onload called: %d - oldDIsplayList has %d elements",
		getTargetPath().c_str(), m_current_frame,
		frame_count, m_on_event_load_called, oldDisplayList.size());
#endif

	// Update current and next frames.
	if (m_play_state == PLAY)
	{
#ifdef GNASH_DEBUG
		log_msg("sprite_instance::advance_sprite we're in PLAY mode");
#endif

		int prev_frame = m_current_frame;

		if (m_on_event_load_called)
		{
#ifdef GNASH_DEBUG
			log_msg(" on_event_load called, incrementing");
#endif
			increment_frame_and_check_for_loop();
#ifdef GNASH_DEBUG
			log_msg(" after increment we are at frame %u/%u", m_current_frame, frame_count);
#endif
		}

		// Execute the current frame's tags.
		// First time execute_frame_tags(0) executed in dlist.cpp(child) or movie_def_impl(root)
		if (m_current_frame != (size_t)prev_frame)
		{
			execute_frame_tags(m_current_frame, TAG_DLIST|TAG_ACTION);
		}
	}
#ifdef GNASH_DEBUG
	else
	{
		log_msg("sprite_instance::advance_sprite we're in STOP mode");
		// shouldn't we execute frame tags anyway when in STOP mode ?
		//execute_frame_tags(m_current_frame);
	}
#endif

	// Advance DisplayList elements which has not been just-added.
	// 
	// These are elements in oldDisplayList cleared of all but elements
	// still in current DisplayList (the other must have been removed
	// by RemoveObject tags). I'm not sure we should do this actually,
	// as maybe we need to execute actions in removed objects *before*
	// we drop them...
	//
	// We do *not* dispatch UNLOAD event on the removed objects here
	// as we assume the event was dispatched by execution of the
	// RemoveObject tag itself. I might be wrong though, will come
	// back to this issue later.
	//
	// Note that we work on a *copy* of oldDisplayList as we're going
	// to need oldDisplayList again later, to extract the list of
	// newly added characters
	//
	DisplayList stillAlive = oldDisplayList;
	stillAlive.clear_except(m_display_list, false);
	//log_msg("Advancing %d pre-existing childs of %s", stillAlive.size(), getTargetPath().c_str());
	stillAlive.advance(delta_time);
	
	// Now execute actions on this timeline, after actions
	// in old childs timelines have been executed.
	//log_msg("Executing actions in %s timeline", getTargetPath().c_str());
	do_actions();

	// Call UNLOAD event of just removed chars !
	DisplayList justRemoved = oldDisplayList;
	justRemoved.clear_except(m_display_list, true);

	// Finally, execute actions in newly added childs
	//
	// These are elements in the current DisplayList, cleared
	// by all elements in oldDisplayList.
	//
	// Of course we do NOT call UNLOAD events here, as
	// the chars we're clearing have *not* been removed:
	// we're simply doing internal work here...
	//
	DisplayList newlyAdded = m_display_list;
	//log_msg("%s has %d current childs and %d old childs", getTargetPath().c_str(), m_display_list.size(), oldDisplayList.size());
	newlyAdded.clear(oldDisplayList, false);
	//log_msg("Advancing %d newly-added (after clearing) childs of %s", newlyAdded.size(), getTargetPath().c_str());
	newlyAdded.advance(delta_time);

	// goto_frame_action (for now) need be executed
	// *after* actions in child sprites have
	// been executed. When action execution order
	// is fixed, we can move this code inside
	// do_action, or completely drop m_goto_frame_action_list
	// or rather push actions in goto_frame target frame
	// directly in the m_action_list (simplest)
	if ( ! m_goto_frame_action_list.empty() )
	{
		IF_VERBOSE_ACTION(
		log_action(" Executing %u actions in "
			"goto_frame_action_list",
			m_goto_frame_action_list.size());
		);
		execute_actions(m_goto_frame_action_list);
		assert(m_goto_frame_action_list.empty());
	}
	
	// Remember current state of the DisplayList for next iteration
	oldDisplayList = m_display_list;
}

// child movieclip advance
void sprite_instance::advance(float delta_time)
{
//	GNASH_REPORT_FUNCTION;

	// child movieclip frame rate is the same the root movieclip frame rate
	// that's why it is not needed to analyze 'm_time_remainder' 
	if (m_on_event_load_called == false)
	{
#ifdef GNASH_DEBUG
		log_msg("Calling ONLOAD event");
#endif
		on_event(event_id::LOAD);	// clip onload

		if (m_has_keypress_event)
		{
			_vm.getRoot().add_keypress_listener(this);
		}
		// Mouse events listening is done in has_mouse_event directly.
		// This shows to work better for attachMovieTest.swf,
		// but might actually be a completely unrelated issue.
		// In particular, copying event handlers in attachMovie should
		// be more closely inspected (and need more ad-hoc testcases)
		//if (m_has_mouse_event)
		//{
			//_vm.getRoot().add_mouse_listener(this);
		//}
	}
	
	advance_sprite(delta_time);

	m_on_event_load_called = true;
}

void
sprite_instance::execute_action(action_buffer& ab)
{
	as_environment& env = m_as_environment; // just type less

	// Do not cleanup locals here, as there's nothing like
	// a movie-frame local scope...
	
	//int local_stack_top = env.get_local_frame_top();
	//env.add_frame_barrier();

	ActionExec exec(ab, env);
	exec();

	//env.set_local_frame_top(local_stack_top);
}

// 0-based frame number !
void
sprite_instance::execute_frame_tags(size_t frame, int typeflags)
{
	testInvariant();

	assert(frame < m_def->get_frame_count());

	if ( frame == 0 && has_looped() )
	{
		// Resort frame0 DisplayList as depth of
		// characters in it might have been
		// externally changed.
		_frame0_chars.sort();

		// Add script objects in current DisplayList
		std::vector<character*> charsToAdd; 
		std::vector<character*> charsToKeep; 
		ScriptObjectsFinder scriptObjFinder(charsToAdd, charsToKeep);
		m_display_list.visitForward(scriptObjFinder);

		// Remove characters which have been removed
		_frame0_chars.clear_except(charsToKeep);

		// NOTE: script objects are *not* allowed to replace depths
		//       of static objects (change second argument to switch)
		_frame0_chars.addAll(charsToAdd, false);
		
		if ( ! (m_display_list == _frame0_chars) ) {
	
			// Set this character as invalidated *before*
			// actually updating the displaylist !
			set_invalidated();
	
			m_display_list = _frame0_chars;			
		};
	}

	// Execute this frame's init actions, if necessary.
	if (m_init_actions_executed[frame] == false)
	{

		const PlayList* init_actions = m_def->get_init_actions(frame);

		if ( init_actions && ! init_actions->empty() )
		{

			IF_VERBOSE_ACTION(
				// Use 1-based frame numbers
				log_action("Executing " SIZET_FMT 
					" *init* tags in frame " SIZET_FMT
					"/" SIZET_FMT
				        " of sprite %s", init_actions->size(),
					frame+1, get_frame_count(),
					getTargetPath().c_str());
			);


			// Need to execute these actions.
			std::for_each(init_actions->begin(), init_actions->end(),
				std::bind2nd(std::mem_fun(&execute_tag::execute), this));

			// Mark this frame done, so we never execute these
			// init actions again.
			m_init_actions_executed[frame] = true;
		}
	}

	const PlayList& playlist = m_def->get_playlist(frame);

	IF_VERBOSE_ACTION(
		// Use 1-based frame numbers
		log_action("Executing " SIZET_FMT " tags in frame "
			SIZET_FMT "/" SIZET_FMT " of sprite %s",
			playlist.size(), frame+1, get_frame_count(),
			getTargetPath().c_str());
	);

	for (PlayList::const_iterator it=playlist.begin(), itEnd=playlist.end();
			it != itEnd; ++it)
	{
		execute_tag* tag = *it;
		if ( typeflags & TAG_DLIST ) tag->execute_state(this);
		if ( typeflags & TAG_ACTION ) tag->execute_action(this);
	}

	if ( frame == 0 && ! has_looped() )
	{
		// Save DisplayList state
		_frame0_chars = m_display_list;
	}

	testInvariant();
}

void sprite_instance::execute_frame_tags_reverse(size_t frame)
{
	testInvariant();

	assert(frame < m_def->get_frame_count());

	const PlayList& playlist = m_def->get_playlist(frame);

	for (unsigned int i=0, n=playlist.size(); i<n; ++i)
	{
	    execute_tag*	e = playlist[i];
	    e->execute_state_reverse(this, frame);
	}

	testInvariant();
}

void sprite_instance::execute_remove_tags(int frame)
{
	    assert(frame >= 0);
	    assert((size_t)frame < m_def->get_frame_count());

	    const PlayList& playlist = m_def->get_playlist(frame);
	    for (unsigned int i = 0; i < playlist.size(); i++)
		{
		    execute_tag*	e = playlist[i];
		    if (e->is_remove_tag())
			{
			    e->execute_state(this);
			}
		}
}

execute_tag*
sprite_instance::find_previous_replace_or_add_tag(int frame,
		int depth, int id)
{
	uint32 depth_id = ((depth & 0x0FFFF) << 16) | (id & 0x0FFFF);

	for (int f = frame - 1; f >= 0; f--)
	{
	    const PlayList& playlist = m_def->get_playlist(f);
	    for (int i = playlist.size() - 1; i >= 0; i--)
		{
		    execute_tag*	e = playlist[i];
		    if (e->get_depth_id_of_replace_or_add_tag() == depth_id)
			{
			    return e;
			}
		}
	}

    return NULL;
}

void
sprite_instance::goto_frame(size_t target_frame_number)
{
//	IF_VERBOSE_DEBUG(log_msg("sprite::goto_frame(%d)\n", target_frame_number));//xxxxx

	//	target_frame_number = iclamp(target_frame_number, 0, m_def->get_frame_count() - 1);
	// Macromedia Flash ignores goto_frame(bad_frame)
	if (target_frame_number > m_def->get_frame_count() - 1 ||
			target_frame_number == m_current_frame)	// to prevent infinitive recursion
	{
		//FIXME: Don't set play state to STOP, just return will be more correct. 
		//  m_current_frame will be incremented in next advance_sprite, so I think 
		//  there will be no 'infinitive recursion' (Zou)
		set_play_state(STOP);
		return;
	}

	// Unless the target frame is the next one, stop playback of soundstream
	int stream_id = get_sound_stream_id();
	if (target_frame_number != m_current_frame+1 && stream_id != -1) {
		sound_handler* sh = get_sound_handler();
		if (sh != NULL) sh->stop_sound(stream_id);
		set_sound_stream_id(-1);
	}

	size_t loaded_frames = get_loaded_frames();
	// target_frame_number is 0-based, get_loaded_frames() is 1-based
	// so in order to goto_frame(3) loaded_frames must be at least 4
	// if goto_frame(4) is called, and loaded_frames is 4 we're jumping
	// forward
	if ( target_frame_number >= loaded_frames )
	{
		IF_VERBOSE_ASCODING_ERRORS(
			log_aserror("GotoFrame(" SIZET_FMT ") targets a yet "
				"to be loaded frame (" SIZET_FMT ") loaded). "
				"We'll wait for it but a more correct form "
				"is explicitly using WaitForFrame instead.",
				target_frame_number+1,
				loaded_frames);

		);
		m_def->ensure_frame_loaded(target_frame_number+1);
	}


	//
	// Construct the DisplayList of the target frame
	//

	if (target_frame_number < m_current_frame)
	// Go backward to a previous frame
	{
		for (size_t f = m_current_frame; f>target_frame_number; --f)
		{
			// CHECK: will this execute "state" tags only ?
			execute_frame_tags_reverse(f);
		}
	}
	else
	// Go forward to a later frame
	{
		// We'd immediately return if target_frame_number == m_current_frame
		assert(target_frame_number > m_current_frame);

		// Construct the DisplayList of the target frame
		for (size_t f = m_current_frame+1; f<target_frame_number; ++f)
		{
			// Second argument requests that only "DisplayList" tags
			// are executed. This means NO actions will be
			// pushed on m_action_list.
			execute_frame_tags(f, TAG_DLIST);
		}

	}

	/// Backup current action list, as we're going to use it
	/// to fetch actions in the target frame
	ActionList actionListBackup = m_action_list;

	// m_action_list contains actions from frame 'target_frame_number'
	// to frame 'm_current_frame', too much than needed, clear it first.
	m_action_list.clear();

	// Get the actions of target frame.(We don't have a direct way to
	// do this, so use execute_frame_tags instead).
	execute_frame_tags(target_frame_number, TAG_ACTION|TAG_DLIST);

	//FIXME: set m_current_frame to the target frame;
	//  I think it's too early to do it here! Later actions in the 
	//  current frame should also be executed(Zou)
	m_current_frame = target_frame_number;      

	// goto_frame stops by default.
	// Zou: ActionGotoFrame tells the movieClip to go to the target frame 
	//  and stop at that frame. 
	set_play_state(STOP);

	// After entering to advance_sprite() m_current_frame points to frame
	// that already is executed. 
	// Macromedia Flash do goto_frame then run actions from this frame.
	// We do too.

	// Stores the target frame actions to m_goto_frame_action_list, which
	// will be executed in after advancing child characters.
	// When actions execution order will be fixed all of this might
	// become obsoleted and the target frame actions could be pushed
	// directly on the preexisting m_action_list (and no need to back it up).
	//
	m_goto_frame_action_list = m_action_list; 

	// Restore ActionList  from backup
	m_action_list = actionListBackup;

}

bool sprite_instance::goto_labeled_frame(const char* label)
{
	size_t target_frame;
	if (m_def->get_labeled_frame(label, &target_frame))
	{
		goto_frame(target_frame);
		return true;
	}
	else
	{
		IF_VERBOSE_MALFORMED_SWF(
		log_swferror("sprite_instance::goto_labeled_frame('%s') "
			"unknown label", label);
		);
		return false;
	}
}

void sprite_instance::display()
{
	//GNASH_REPORT_FUNCTION;

	if (get_visible() == false)
	{
		// We're invisible, so don't display!
		
		// Note: dlist.cpp will avoid to even call display() so this will probably
		// never happen.
		return;
	}
	
	// check if the sprite (and it's childs) needs to be drawn
	InvalidatedRanges ranges;
	m_display_list.add_invalidated_bounds(ranges, true);
	
	// expand to bounds of _drawable
	rect drawable_bounds; 
	drawable_bounds.expand_to_transformed_rect(get_world_matrix(), _drawable->get_bound());
	ranges.add(drawable_bounds.getRange());
	
	if (gnash::render::bounds_in_clipping_area(ranges))
	{
		_drawable->finalize();
		// TODO: I'd like to draw the definition directly..
		//       but it seems that the backend insists in
		//       accessing the *parent* of the character
		//       passed as "instance" for the drawing.
		//       When displaying top-level movie this will
		//       be NULL and gnash will segfault
		//       Thus, this drawable_instance is basically just
		//       a container for a parent :(
		_drawable_inst->display();

		m_display_list.display();
		clear_invalidated();
	}
	  
	do_display_callback();
}

void sprite_instance::swap_characters(character* ch1, character* ch2)
{
	ch1->set_invalidated();
	ch2->set_invalidated();

	m_display_list.swap_characters(ch1, ch2);	
}

bool
sprite_instance::attachCharacter(character& newch, uint16_t depth)
{

	// place_character() will set depth on newch
	m_display_list.place_character(
		&newch,
		depth,
		cxform(),
		matrix(),
		1.0,
		0);

	return true; // FIXME: check return from place_character above ?
}

character*
sprite_instance::add_display_object(
		uint16_t character_id,
		const char* name,
		const std::vector<swf_event*>& event_handlers,
		uint16_t depth, 
		bool replace_if_depth_is_occupied,
		const cxform& color_transform, const matrix& matrix,
		float ratio, uint16_t clip_depth)
{
//	    GNASH_REPORT_FUNCTION;
	assert(m_def != NULL);

	character_def*	cdef = m_def->get_character_def(character_id);
	if (cdef == NULL)
	{
		IF_VERBOSE_MALFORMED_SWF(
			log_swferror("sprite_instance::add_display_object(): "
				"unknown cid = %d", character_id);
		);
		return NULL;
	}

	character* existing_char = m_display_list.get_character_at_depth(depth);

	if (existing_char)
	{

		// If we already have this object on this
		// plane, then move it instead of replacing it.
		if ( existing_char->get_id() == character_id )
		{
			// TODO: update name ?
			move_display_object(depth, true, color_transform,
				true, matrix, ratio, clip_depth);
			return NULL;
		}

		// If we've been asked NOT to replace existing chars just
		// return NULL now
		if ( ! replace_if_depth_is_occupied )
		{
			return NULL;
		}
	}
	//printf("%s: character %s, id is %d, count is %d\n", __FUNCTION__, existing_char->get_name(), character_id,m_display_list.get_character_count()); 

	assert(cdef);
	boost::intrusive_ptr<character> ch = cdef->create_character_instance(this, character_id);
	assert(ch.get() != NULL);

	// Syntetize an instance name if this character doesn't have one
	// TODO: check if we need to do this *only* for sprite characters
	//       also, consider asking for "nextInstanceName" to the character
	//       definition...
	std::string instance_name;
	if ( name ) instance_name = name;
	else instance_name = getNextUnnamedInstanceName();

	ch->set_name(instance_name.c_str());

	// Attach event handlers (if any).
	for (size_t i = 0, n = event_handlers.size(); i < n; i++)
	{
		event_handlers[i]->attach_to(ch.get());
	}

	m_display_list.place_character(
		ch.get(),
		depth,
		color_transform,
		matrix,
		ratio,
		clip_depth);

	assert(ch == NULL || ch->get_ref_count() > 1);
	return ch.get();
}

void
sprite_instance::replace_display_object(
		uint16_t character_id,
		const char* name,
		uint16_t depth,
		bool use_cxform,
		const cxform& color_transform,
		bool use_matrix,
		const matrix& mat,
		float ratio,
		uint16_t clip_depth)
{
	assert(m_def != NULL);
	// printf("%s: character %s, id is %d\n", __FUNCTION__, name, character_id); // FIXME: debugging crap

	character_def*	cdef = m_def->get_character_def(character_id);
	if (cdef == NULL)
	{
		log_error("sprite::replace_display_object(): "
			"unknown cid = %d\n", character_id);
		return;
	}
	assert(cdef);

	boost::intrusive_ptr<character> ch = cdef->create_character_instance(this,
			character_id);

	replace_display_object(
		ch.get(), name, depth,
		use_cxform, color_transform,
		use_matrix, mat,
		ratio, clip_depth);
}

void sprite_instance::replace_display_object(
		character* ch,
		const char* name,
		uint16_t depth,
		bool use_cxform,
		const cxform& color_transform,
		bool use_matrix,
		const matrix& mat,
		float ratio,
		uint16_t clip_depth)
{
    //printf("%s: character %s, id is %d\n", __FUNCTION__, name, ch->get_id()); // FIXME:

    assert(ch != NULL);

    if (name != NULL && name[0] != 0)
	{
	    ch->set_name(name);
	}

	 set_invalidated();

    m_display_list.replace_character(
	ch,
	depth,
	use_cxform,
	color_transform,
	use_matrix,
	mat,
	ratio,
	clip_depth);
	
}

int sprite_instance::get_id_at_depth(int depth)
{
    character*	ch = m_display_list.get_character_at_depth(depth);
    if ( ! ch ) return -1;
    else return ch->get_id();
}

void sprite_instance::increment_frame_and_check_for_loop()
{
	//GNASH_REPORT_FUNCTION;

	size_t frame_count = m_def->get_frame_count();
	if ( ++m_current_frame >= frame_count )
	{
		// Loop.
		m_current_frame = 0;
		m_has_looped = true;
		if (frame_count > 1)
		{
			//m_display_list.reset();
		}
	}

#if 0 // debugging
	log_msg("Frame %u/%u, bytes %u/%u",
		m_current_frame, frame_count,
		get_bytes_loaded(), get_bytes_total());
#endif
}

/// Find a character hit by the given coordinates.
class MouseEntityFinder {

	character* _m;

	float _x;

	float _y;

public:

	MouseEntityFinder(float x, float y)
		:
		_m(NULL),
		_x(x),
		_y(y)
	{}

	bool operator() (character* ch)
	{
		if ( ! ch->get_visible() ) return true;

		character* te = ch->get_topmost_mouse_entity(_x, _y);
		if ( te )
		{
			_m = te;
			return false; // done
		}

		return true; // haven't found it yet
	}

	character* getEntity() { return _m; }
		
};

character*
sprite_instance::get_topmost_mouse_entity(float x, float y)
{
	//GNASH_REPORT_FUNCTION;

	if (get_visible() == false)
	{
		return NULL;
	}

	matrix	m = get_matrix();
	point	p;
	m.transform_by_inverse(&p, point(x, y));

	MouseEntityFinder finder(p.m_x, p.m_y);
	m_display_list.visitBackward(finder);
	character* ch = finder.getEntity();
	if ( ! ch ) 
	{
		ch = _drawable_inst->get_topmost_mouse_entity(p.m_x, p.m_y);
	}

	if ( ch && can_handle_mouse_event() )
	{
		return this;
	}
	else
	{
		return ch; // might be NULL
	}

}

bool
sprite_instance::can_handle_mouse_event() const
{
	as_value dummy;

	// Event handlers that qualify as mouse event handlers.
	static const event_id EH[] =
	{
		event_id(event_id::PRESS),
		event_id(event_id::RELEASE),
		event_id(event_id::RELEASE_OUTSIDE),
		event_id(event_id::ROLL_OVER),
		event_id(event_id::ROLL_OUT),
		event_id(event_id::DRAG_OVER),
		event_id(event_id::DRAG_OUT),
		// MOUSE_{DOWN,UP,MOVE} are handled
		// differently, in that they are called
		// reguardles of mouse position
		// See has_mouse_event
		//event_id(event_id::MOUSE_DOWN),
		//event_id(event_id::MOUSE_UP)
		//event_id(event_id::MOUSE_MOVE)
	};

	int swfversion =  _vm.getSWFVersion();
	for (unsigned int i = 0; i < ARRAYSIZE(EH); i++)
	{
		const event_id &event = EH[i];

		// Check event handlers
		if (get_event_handler(event.id(), &dummy))
		{
			return true;
		}

		// Check user-defined event handlers
		// TODO: check if it's possible to actually
		//       have an hard-coded handler and a user-defined
		//       one. If this is not the case we should add
		//       gettersetter memebers for all these handlers.
		//
		std::string fname = event.get_function_name();
		if ( swfversion < 7 )
		{
			// TODO: have event.get_function_name()
			//       return an SWF-contextual string  instead!
			boost::to_lower(fname, _vm.getLocale());
		}

		// The const_cast is needed because get_member, due to
		// possible "getter" methods executing stuff, is a non-const
		// function. We take the "risk" here...
		if (const_cast<sprite_instance*>(this)->get_member(fname, &dummy))
		{
			return true;
		}
	}

	return false;
}
		
void sprite_instance::restart()
{
    m_current_frame = 0;
    m_update_frame = true;
    m_has_looped = false;
    m_play_state = PLAY;

    execute_frame_tags(m_current_frame);
    //we don't have the concept of a DisplayList update anymore
    //m_display_list.update();
}

float sprite_instance::get_height() const
{
	HeightFinder f;
	// the const_cast is just to avoid defining a const version
	// of DisplayList::visitForward, HeightFinder will NOT
	// modify the DisplayList elements in any way
	const_cast<DisplayList&>(m_display_list).visitForward(f);
	return f.getHeight(); 
}

float sprite_instance::get_width() const
{
	WidthFinder f;
	// the const_cast is just to avoid defining a const version
	// of DisplayList::visitForward, WidthFinder will NOT
	// modify the DisplayList elements in any way
	const_cast<DisplayList&>(m_display_list).visitForward(f);
	return f.getWidth(); 
}

character*
sprite_instance::get_character(int /* character_id */)
{
	//return m_def->get_character_def(character_id);
	// @@ TODO -- look through our dlist for a match
	log_msg("FIXME: %s doesn't even check for a char",
		__PRETTY_FUNCTION__);
	return NULL;
}

//float
//sprite_instance::get_timer() const
//{
//	return m_root->get_timer();
//}


sprite_instance*
sprite_instance::get_root_movie()
{
	assert(m_root);
	return m_root; // could as well be myself !
}

float
sprite_instance::get_pixel_scale() const
{
	return _vm.getRoot().get_pixel_scale();
}

void
sprite_instance::get_mouse_state(int& x, int& y, int& buttons)
{
	_vm.getRoot().get_mouse_state(x, y, buttons);
}

void
sprite_instance::stop_drag()
{
	assert(m_parent == NULL);	// we must be the root movie!!!
	_vm.getRoot().stop_drag();
}

float
sprite_instance::get_background_alpha() const
{
    // @@ this doesn't seem right...
    return _vm.getRoot().get_background_alpha();
}

void
sprite_instance::set_background_color(const rgba& color)
{
	_vm.getRoot().set_background_color(color);
}


/* public */
void
sprite_instance::set_textfield_variable(const std::string& name,
		edit_text_character* ch)
{
	assert(ch);

	// lazy allocation
	if ( ! _text_variables.get() )
	{
		_text_variables.reset(new TextfieldMap);
	}

	// TODO: should variable name be considered case-insensitive ?
	_text_variables->operator[] (name) = ch;
}

/* private */
edit_text_character*
sprite_instance::get_textfield_variable(const std::string& name)
{
	// nothing allocated yet...
	if ( ! _text_variables.get() ) return NULL;

	// TODO: should variable name be considered case-insensitive ?
	TextfieldMap::iterator it = _text_variables->find(name);
	if ( it == _text_variables->end() )
	{
		return NULL;
	}
	else
	{
		return it->second.get();
	}
} 


void 
sprite_instance::add_invalidated_bounds(InvalidatedRanges& ranges, 
	bool force)
{

	// nothing to do if this sprite is not visible
	if (!m_visible || get_cxform().is_invisible() )
	{
    ranges.add(m_old_invalidated_ranges); // (in case we just hided)
		return;
	}

	if ( ! m_invalidated && ! m_child_invalidated && ! force )
	{
		return;
	}
  
 
  // m_child_invalidated does not require our own bounds
  if ( m_invalidated || force )      
  {
  	// Add old invalidated bounds
		ranges.add(m_old_invalidated_ranges); 
  }
  
  
	m_display_list.add_invalidated_bounds(ranges, force||m_invalidated);

	_drawable_inst->add_invalidated_bounds(ranges, force||m_invalidated);

}


const char*
sprite_instance::call_method_args(const char* method_name,
		const char* method_arg_fmt, va_list args)
{
    // Keep m_as_environment alive during any method calls!
    boost::intrusive_ptr<as_object>	this_ptr(this);

    return call_method_parsed(&m_as_environment, this,
		method_name, method_arg_fmt, args);
}

const std::string&
sprite_instance::getTargetPath() const
{
	if ( _target.empty() ) _target = computeTargetPath();
	else
	{
		// What if set_name() is called *after*
		// we computed the _target string ?
		// let's check this... the design doesn't
		// take the problem into account..
		assert (_target == computeTargetPath() );
	}

	return _target;
}

/*private*/
std::string
sprite_instance::computeTargetPath() const
{

	// TODO: check what happens when this character
	//       is a movie_instance loaded into another
	//       running movie.

	typedef std::vector<std::string> Path;
	Path path;

	// Build parents stack
	const character* ch = this;
	for (;;)
	{
		const character* parent = ch->get_parent();

		// Don't push the _root name on the stack
		if ( ! parent )
		{
			assert(ch->get_name().empty());
			break;
		}

		path.push_back(ch->get_name());
		ch = parent;
	} 

	if ( path.empty() ) return "/";

	// Build the target string from the parents stack
	std::string target;
	for ( Path::reverse_iterator
			it=path.rbegin(), itEnd=path.rend();
			it != itEnd;
			++it )
	{
		target += "/" + *it; 
	}

	return target;
}

const char*
sprite_instance::get_text_value() const
{
	if ( ! _target_dot.empty() ) return _target_dot.c_str();

	_target_dot = "_level0" + getTargetPath();
	for (std::string::size_type i=0; i<_target_dot.length(); ++i)
	{
		if ( _target_dot[i] == '/' ) _target_dot[i] = '.';
	}
	return _target_dot.c_str();
}

// WARNING: THIS SNIPPET NEEDS THE CHARACTER TO BE "INSTANTIATED", which is
//          it's target path needs to exist, or any as_value for it will be
//          a dangling reference to an unexistent sprite !
void
sprite_instance::construct()
{
#ifdef GNASH_DEBUG
	log_msg("Constructing sprite '%s'", getTargetPath().c_str());
#endif

	// Backup the DisplayList *before* manipulating it !
	assert( oldDisplayList.empty() );

	on_event(event_id::CONSTRUCT);
	execute_frame_tags(0, TAG_DLIST|TAG_ACTION);	

	sprite_definition* def = dynamic_cast<sprite_definition*>(m_def.get());

	// We won't "construct" top-level movies
	if ( ! def ) return;

	// instance name will be needed for properly setting up
	// a reference to 'this' object.
	assert(!_name.empty());

	as_function* ctor = def->getRegisteredClass();
	//log_msg("Attached sprite's registered class is %p", (void*)ctor); 

	// TODO: builtin constructors are different from user-defined ones
	// we should likely change that. See also vm/ASHandlers.cpp (construct_object)
	if ( ctor && ! ctor->isBuiltin() )
	{
		// Set the new prototype *after* the constructor was called
		as_object* proto = ctor->getPrototype();
		set_prototype(proto);

		//log_msg("Calling the user-defined constructor against this sprite_instance");
		as_value ret; // we don't use the constructor return (should we?)
		fn_call call(&ret, this, &(get_environment()), 0, 0);
		(*ctor)(call);

	}
}

void
sprite_instance::set_name(const char* name)
{
	_name = name;

	// Reset these so they get computed next
	// time someone request them.
	_target.clear();
	_target_dot.clear();
}

/*private static*/
std::string
sprite_instance::getNextUnnamedInstanceName()
{
	std::stringstream ss;
	ss << "instance" << ++_lastUnnamedInstanceNum;
	return ss.str();
}

/* private static */
bool
sprite_instance::sameEvents(const Events& eventsMap, const SWFEventsVector& eventsVect)
{
	size_t n = eventsVect.size();
	if (eventsMap.size() != n) return false;

	for (size_t i = 0; i < n; i++)
	{
		Events::const_iterator it = eventsMap.find(eventsVect[i]->m_event);

		if ( it == eventsMap.end() ) return false;

		as_value result = it->second;
		// compare actionscipt in event
		if (eventsVect[i]->m_method != result) return false;
	}

	return true;
	
}

bool
sprite_instance::loadMovie(const URL& url)
{
	boost::intrusive_ptr<movie_definition> md ( create_library_movie(url) );
	if (md == NULL)
	{
		log_error("can't create movie_definition for %s\n",
			url.str().c_str());
		return false;
	}

	boost::intrusive_ptr<movie_instance> extern_movie;
	extern_movie = md->create_movie_instance();
	if (extern_movie == NULL)
	{
		log_error("can't create extern movie_instance "
			"for %s\n", url.str().c_str());
		return false;
	}

	save_extern_movie(extern_movie.get());

	const char* name = get_name().c_str();
	uint16_t depth = get_depth();
	bool use_cxform = false;
	cxform color_transform = get_cxform();
	bool use_matrix = false;
	matrix mat = get_matrix();
	float ratio = get_ratio();
	uint16_t clip_depth = get_clip_depth();
	//character* new_movie = extern_movie->get_root_movie();

	// Get a pointer to our own parent 
	character* parent = get_parent();
	if ( parent )
	{
		extern_movie->set_parent(parent);

		sprite_instance* parent_sp = parent->to_movie();
		assert(parent_sp);
		parent_sp->replace_display_object(
				   extern_movie.get(),
				   name,
				   depth,
				   use_cxform,
				   color_transform,
				   use_matrix,
				   mat,
				   ratio,
				   clip_depth);
	}
	else
	{
		movie_root& root = _vm.getRoot();
		// Make sure we won't kill ourself !
		assert(get_ref_count() > 1);
		root.setRootMovie(extern_movie.get());
	}

	return true;
}

void
sprite_instance::has_mouse_event()
{
	m_has_mouse_event = true;
	_vm.getRoot().add_mouse_listener(this);
}

} // namespace gnash
